"""Plugin for finding possible threat actors based on sentences"""

import addict
from act.scio.plugin import BasePlugin, Result
from typing import Text, List, Set
import nltk
import nltk.stem


class Plugin(BasePlugin):
    """Plugin to detect threat actors"""

    name = "nlp_actors"
    info = "Extract actors based on language from a body of text"
    version = "0.1"
    dependencies: List[Text] = ["pos_tag"]

    async def analyze(self, nlpdata: addict.Dict) -> Result:

        res = addict.Dict()

        threat_stem_postfix = {
            'threat',     # threat
            'crimin',     # criminal, criminals
            'crime',      # crime
            'espionage',  # espionage
            'hack',       # hack, hacking,
            'hacker',     # hacker, hackers
            'crack',      # cracking, crack
            'cracker',    # cracker, crackers
            'adversari',  # adversary, adversaries
            'terrorist'   # terrorist, terrorists
        }

        group_stem_postfix = {
            'group',  # group, groups
            'actor',  # actor, actors
            'unit',   # unit, untis
            'agent',  # agent, agents
            'organ'   # organization, organizations
        }

        false_positive_filter = ["top",
                                 "unknown",
                                 "cyber"]  # "top threat groups", "cyber threat actors" etc...

        possible_ta_tag_types = {"NNP", "NNPS", "NN", "NNS"}
        possible_tag_types = {"NNP", "NNPS", "NN", "NNS", "JJ", "JJS"}
        chain_tags = {",", ":", "CC"}
        lookbefore_tags: Set[Text] = set()
        lookbefore_tags.update(chain_tags)
        lookbefore_tags.update(possible_tag_types)

        ps = nltk.stem.PorterStemmer()

        first_stage_found = False

        pos_actors: List[Text] = []
        # Look through all tokens. If any token relating to a threat actors is
        # found, look-before and collect all nouns while the tokens are nouns
        # or part of a listing.
        for i, (token, tag) in enumerate(nlpdata.pos_tag.tokens):
            if first_stage_found and tag in possible_tag_types and ps.stem(token) in group_stem_postfix:
                if nlpdata.pos_tag.tokens[i - 2][1] not in possible_tag_types:
                    first_stage_found = False
                    continue
                n = i - 1
                while nlpdata.pos_tag.tokens[n][1] in lookbefore_tags:
                    n -= 1

                current_actor: List[Text] = []
                for (subtoken, pos_tag) in nlpdata.pos_tag.tokens[n:i-1]:
                    if pos_tag in chain_tags:
                        if current_actor:
                            pos_actors.append(" ".join(current_actor))
                            current_actor = []
                    elif pos_tag in possible_ta_tag_types:
                        if subtoken in false_positive_filter:
                            continue
                        current_actor.append(subtoken)
                if current_actor:
                    pos_actors.append(" ".join(current_actor))

            first_stage_found = bool(tag in possible_tag_types and
                                     ps.stem(token) in threat_stem_postfix)

        res.actors = pos_actors
        return Result(name=self.name, version=self.version, result=res)



"""Copyright 2020 mnemonic AS <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for
any purpose with or without fee is hereby granted, provided that the
above copyright notice and this permission notice appear in all
copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

---
SCIO API
"""

import argparse
import base64
import hashlib
import logging
import os
import re
from functools import lru_cache
from typing import Dict, Optional, Text, Union

import act.scio.config
import act.scio.es
import caep
import elasticsearch
import greenstalk  # type: ignore
import uvicorn
from fastapi import Depends, FastAPI, HTTPException, Response
from fastapi.responses import FileResponse, PlainTextResponse
from pydantic import BaseModel, StrictInt, StrictStr
from pydantic.types import constr

XDG_CACHE = os.path.expanduser(os.environ.get("XDG_CACHE_HOME", "~/.cache"))

app = FastAPI()

# pylint: disable=too-few-public-methods


class Document(BaseModel):
    """ Document model """
    content: StrictStr
    filename: StrictStr


class LookupResponse(BaseModel):
    """ Response model for document search """
    filename: StrictStr
    content_type: StrictStr


class SubmitResponse(BaseModel):
    """ Response model for document submit """
    filename: StrictStr
    hexdigest: StrictStr
    count: StrictInt
    error: Optional[StrictStr]


# parse_args() is executed the first time and later the cached result
# is used. In this way, we use to configure settings in API endpoints
@lru_cache()
def parse_args() -> argparse.Namespace:
    """Helper setting up the argsparse configuration"""

    arg_parser = act.scio.config.parse_args("Scio API")
    arg_parser.add_argument('--port', type=int, default=3000,
                            help="API port to listen on (default=3000)")
    arg_parser.add_argument('--reload', action="store_true",
                            help="Reload web server on file change (dev mode)")
    arg_parser.add_argument('--document-path', default=caep.get_cache_dir("scio/documents"),
                            help=f"Storage path for documents = {XDG_CACHE}/scio/documents")
    arg_parser.add_argument('--host', dest='host', default='127.0.0.1',
                            help="Host interface (default=127.0.0.1)")

    args = caep.config.handle_args(arg_parser, "scio/etc", "scio.ini", "api")

    if not os.path.isdir(args.document_path):
        os.makedirs(args.document_path)
        logging.info("Created directory: %s", args.document_path)

    args.beanstalk_client = act.scio.config.beanstalk_client(args, use="scio_doc")
    args.elasticsearch_client = act.scio.config.elasticsearch_client(args)

    return args


def document_lookup(document_id: Text,
                    elasticsearch_client: elasticsearch.client.Elasticsearch) -> LookupResponse:
    """ Lookup document location and content type from document_id (hexdigest) """

    filename: Text = ""
    content_type: Text = ""

    try:
        res = elasticsearch_client.get(index="scio2", id=document_id.lower())
        filename = res["_source"].get("filename")
        content_type = res["_source"].get("metadata", {}).get("Content-Type")
    except elasticsearch.exceptions.NotFoundError:
        pass

    return LookupResponse(
        filename=filename,
        content_type=content_type
    )


@app.post("/submit")
async def submit(doc: Document, args: argparse.Namespace = Depends(parse_args)) -> SubmitResponse:
    # Depends on parse_args which are used for settings. The result
    # is cached the first time it is executed
    """ Submit document """
    filename = os.path.join(args.document_path, os.path.basename(doc.filename))
    content: bytes = base64.b64decode(doc.content)
    with open(filename, "bw") as f:
        f.write(content)

    response = SubmitResponse(
        filename=filename,
        hexdigest=hashlib.sha256(content).hexdigest(),
        count=len(content),
        error=None)

    args.beanstalk_client.put(response.json().encode("utf8"))
    return response


@ app.get("/indicators/{indicator_type}", response_class=PlainTextResponse)
def indicators(indicator_type: constr(regex=r"^(ipv4|ipv6|uri|email|fqdn|md5|sha1|sha256)$"),
             last: constr(regex=r'^\d+[yMwdhms]?$') = "90d",
             args: argparse.Namespace = Depends(parse_args)) -> PlainTextResponse:
    """ Download indicators

    Allowed indicator types:
    * ipv4
    * ipv6
    * uri
    * email
    * fqdn
    * md5
    * sha1
    * sha256

    You can also specify the maximum age of the document you want to
    get indicators from with the `last` argument (default=90d). The
    format should be either be <NUM><TIME UNIT>, where TIME UNIT can be one of:

    y (year)
    M (month)
    w (week)
    d (day)
    h (hour)
    m (minute)
    s (second)

    OR <EPOC> (only digits) where the EPOC is a unix timestamp in milliseconds

    """

    if not args.elasticsearch_client:
        raise HTTPException(status_code=412, detail="Elasticsearch is not configured")

    if re.search(r"^\d+$", last):
        # Only digits - assume unix timestamp
        start = last
    else:
        start = f"now-{last}"

    term = f"indicators.{indicator_type}.keyword"

    res = act.scio.es.aggregation(
        args.elasticsearch_client,
        terms=[term],
        start=start,
        end="now",
    )

    return PlainTextResponse(
            "\n".join(row[0].get(term) for row in res)
    )


@ app.get("/download/{document_id}")
def download(document_id: constr(regex=r"^[0-9A-Fa-f]{64}$"),
             args: argparse.Namespace = Depends(parse_args)) -> Response:
    """ Download document as original content"""
    res = document_lookup(document_id, args.elasticsearch_client)

    if not os.path.isfile(res.filename):
        return Response(content="File not found", media_type="application/text")

    return FileResponse(
        res.filename,
        filename=os.path.basename(res.filename),
        media_type=res.content_type)


@ app.get("/download_json/{document_id}")
def download_json(document_id: constr(regex=r"^[0-9A-Fa-f]{64}$"),
                  args: argparse.Namespace = Depends(parse_args)) -> Union[Response, Dict]:
    """ Download document base64 decoded in json struct """
    res = document_lookup(document_id, args.elasticsearch_client)

    if not os.path.isfile(res.filename):
        return {
            "error": "File not found",
            "bytes": 0,
        }

    content = open(res.filename, "rb").read()

    return {
        "error": None,
        "bytes": len(content),
        "content": base64.b64encode(content),
        "encoding": "base64"
    }


def main() -> None:
    """ Main API loop """
    args = parse_args()

    uvicorn.run(
        "act.scio.api:app",
        host=args.host,
        port=args.port,
        log_level="info",
        reload=args.reload)


if __name__ == "__main__":
    main()



"""All functions related to scio upload"""

from typing import IO, Text, Dict

import base64

import requests


def read_as_base64(obj: IO) -> Text:
    """Create a base64 encoded string from a file like object"""

    encoded_bytes = base64.b64encode(obj.read())
    return encoded_bytes.decode('ascii')


def to_scio_submit_post_data(obj: IO, file_name: Text) -> Dict[Text, Text]:
    """Take a file like object, and return a dictionary on the correct form for
    submitting to the SCIO API (https://github.com/mnemonic-no/act-scio-api)"""

    return {'content': read_as_base64(obj), 'filename': file_name}


def upload(url: Text, filename: Text) -> None:
    """Upload a file to the Scio engine"""

    with open(filename, "rb") as file_h:
        post_data = to_scio_submit_post_data(file_h, filename)

        # Disable all proxy settings from environment
        session = requests.Session()
        session.trust_env = False

        req = session.post(url, json=post_data)
        if req.status_code != 200:
            raise UploadError("Status {0}".format(req.status_code))


class UploadError(Exception):
    """Wrap all errors related to uploading data to scio"""



import addict
from act.scio.vocabulary import Vocabulary
from act.scio.plugin import BasePlugin, Result
from typing import Text, List
import configparser
import nltk
import nltk.stem
import os


class Plugin(BasePlugin):
    name = "sectors"
    info = "Extract sectors from a body of text"
    version = "0.1"
    dependencies: List[Text] = ["pos_tag"]

    async def analyze(self, nlpdata: addict.Dict) -> Result:

        res = addict.Dict()

        sector_stem_postfix = {
            'compani',  # company, companies, [...],
            'industri',  # industry, industries, [...],
            'sector',   # sector, sectors, [...],
            'servic',   # service, services, [...],
            'organ',   # organization, organizations, [...],
            'provid',  # provider, providers, [...],
        }

        posible_tag_types = {"NNP", "NNPS", "NN", "NNS"}
        lookbefore_tags = {",", ":", "CC"}
        lookbefore_tags.update(posible_tag_types)

        ps = nltk.stem.PorterStemmer()

        pos_sectors: List[Text] = []
        # Look through all tokens. If any token relating to a sector is found,
        # look-before and collect all nouns while the tokens are nouns or part
        # of a listing.
        for i, (token, tag) in enumerate(nlpdata.pos_tag.tokens):
            if tag in posible_tag_types and ps.stem(token) in sector_stem_postfix:
                n = i - 1
                while nlpdata.pos_tag.tokens[n][1] in lookbefore_tags:
                    n -= 1
                pos_sectors += [token for (token, pos_tag)
                                in nlpdata.pos_tag.tokens[n:i]
                                if pos_tag in posible_tag_types]

        ini = configparser.ConfigParser()
        ini.read([os.path.join(self.configdir, "sectors.ini")])
        ini['sectors']['alias'] = os.path.join(self.configdir, ini['sectors']['alias'])

        vocab = Vocabulary(ini['sectors'])
        sectors = []
        unknown_sectors = []
        for pos_sector in pos_sectors:
            primary = vocab.get(pos_sector, primary=True)
            if primary:
                sectors.append(primary)
            else:
                unknown_sectors.append(pos_sector)

        res.sectors = sectors
        res.unknown_sectors = unknown_sectors
        return Result(name=self.name, version=self.version, result=res)
